###############################################################################
#
#   Copyright: (c) 2017 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import load_system_configuration, Date

from .utils import encode_message, request_to_uid
from .exceptions import NoDataserverAvailable

import tornado.ioloop
import tornado.web
import tornado.httpserver
import tornado.tcpclient
import tornado.gen

import asyncmc
#import memcache
import socket
import datetime
import random
import logging
import os

RT_TIMEOUT = 300  # in seconds


###############################################################################
class DataRouter(tornado.web.Application):
    # -------------------------------------------------------------------------
    def __init__(self, timeout=60, logger=None, *args, **kwds):
        super().__init__(*args, **kwds)
        self.logger = logger or logging.getLogger(__name__)
        self.timeout = datetime.timedelta(seconds=timeout)

        config = load_system_configuration()
        self.mc_servers = [config.get("memcache", "url")]
        self.dataservers = []

    # -------------------------------------------------------------------------
    def add_dataserver(self, server_id):
        if server_id not in self.dataservers:
            self.dataservers.append(server_id)
            self.logger.info("{0!s} registered "
                "as available server".format(server_id))
            return True
        return False

    # -------------------------------------------------------------------------
    def drop_dataserver(self, server_id, msg):
        self.logger.info("dropping "
            "dataserver {0!s}: {1:s}".format(server_id, msg))
        try:
            self.dataservers.remove(server_id)
        except ValueError:
            pass

    # -------------------------------------------------------------------------
    def select_dataserver(self):
        try:
            return random.choice(self.dataservers)
        except IndexError:
            raise NoDataserverAvailable()

    # -------------------------------------------------------------------------
    def start(self, port=None):
        config = load_system_configuration()
        port = port or config.getint("datafeed", "router_port")

        # --- windows-specific hack: increase the file descriptor limit to
        #     circumvent "too many file descriptors in select()" error
        if os.name == "nt":
            import win32file
            win32file._setmaxstdio(2048)
            assert win32file._getmaxstdio() == 2048

        # --- start HTTP server with an idle_connection_timeout so that idle
        #     connections don't stick around for more than 90 seconds
        http_server = tornado.httpserver.HTTPServer(
            self, idle_connection_timeout=90)
        http_server.listen(port)

        # --- start the memcache client
#        self.cache = memcache.Client(servers=self.mc_servers)
        self.cache = asyncmc.Client(servers=self.mc_servers,
                                    loop=tornado.ioloop.IOLoop.current())

        address = socket.gethostbyname(socket.gethostname())
        self.logger.info("listening on {0!s}:{1!s}".format(address, port))

        try:
            tornado.ioloop.IOLoop.current().start()
        except KeyboardInterrupt:
            tornado.ioloop.IOLoop.current().stop()
        finally:
            self.cleanup()

    # -------------------------------------------------------------------------
    def cleanup(self):
        self.logger.info("shutting down router")

    # -------------------------------------------------------------------------
    @tornado.gen.coroutine
    def with_timeout(self, future):
        generator = yield tornado.gen.with_timeout(self.timeout, future)
        return generator

    # -------------------------------------------------------------------------
    @tornado.gen.coroutine
    def fetch(self, req, addr, port):
        client = tornado.tcpclient.TCPClient()
        stream = yield self.with_timeout(client.connect(addr, port))

        try:
            yield self.with_timeout(stream.write(encode_message(req)))
            response = yield self.with_timeout(stream.read_until(b"\n"))
        finally:
            try:
                stream.close()
            except tornado.iostream.StreamClosedError:
                pass

        return response

    # -------------------------------------------------------------------------
    @tornado.gen.coroutine
    def process_request(self, req, real_time):
        # --- determine the unique request id, used for caching
        req_uid = request_to_uid(req, real_time)

        # --- first try fetching response from cache
        while True:
#            response = self.cache.get(req_uid)
            response = yield self.cache.get(req_uid)
            if response == "CacheLocked":
                yield tornado.gen.sleep(0.25)
            else:
                break

        if response is None:
            # --- no valid data stored in cache, send request to datafeed
            #     router.
            #     NB: we first lock the cache with a timeout and then we only
            #         set the cache if the response is valid.
#            self.cache.set(req_uid, "CacheLocked",
#                           self.timeout.seconds, noreply=True)
            yield self.cache.set(req_uid, "CacheLocked",
                                 self.timeout.seconds, noreply=True)

            while True:
                # --- select dataserver or return error if none is available
                try:
                    addr, port = self.select_dataserver()
                except NoDataserverAvailable:
                    return 503, "Dataservers not available or unresponsive"

                try:
                    response = yield self.fetch(req, addr, port)
                except (tornado.gen.TimeoutError,
                        tornado.iostream.StreamClosedError):
                    # --- connection unresponsive: drop dataserver from list
                    #     and process the request again
                    self.drop_dataserver((addr, port), msg="timeout")
                    continue
                else:
                    break

            if real_time:
                expiry = RT_TIMEOUT
            else:
                expiry = int(Date.today().eod().timestamp())

#            self.cache.set(req_uid, response, expiry, noreply=True)
            yield self.cache.set(req_uid, response, expiry, noreply=True)

            self.logger.debug("returning value from dataserver "
                "(RT={0!s}), req_uid='{1!s}'".format(real_time, req_uid))

        else:
            self.logger.debug("returning value "
                "from cache (RT={0!s})".format(real_time))

        return 200, response


###############################################################################
class RegistrationHandler(tornado.web.RequestHandler):
    # -------------------------------------------------------------------------
    def put(self):
        self.set_header("Access-Control-Allow-Origin", "*")

        addr = self.get_argument("address")
        port = self.get_argument("port")

        added = self.application.add_dataserver((addr, port))
        if added:
            self.set_status(201)
        else:
            self.set_status(205)


###############################################################################
class BbgBDPHandler(tornado.web.RequestHandler):
    # -------------------------------------------------------------------------
    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def get(self):
        self.set_header("Content-Type", "application/json")
        self.set_header("Access-Control-Allow-Origin", "*")

        request = {
            "type": "BDP",
            "sec": self.get_argument("sec"),
            "field": self.get_argument("field"),
            "overrides": self.get_argument("overrides", "null"),
        }

        rt = self.get_argument("RT", False)

        status, resp = yield self.application.process_request(request, rt)

        self.set_status(status, None if status == 200 else resp)
        self.write(resp)


###############################################################################
class BbgBDHHandler(tornado.web.RequestHandler):
    # -------------------------------------------------------------------------
    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def get(self):
        self.set_header("Content-Type", "application/json")
        self.set_header("Access-Control-Allow-Origin", "*")

        request = {
            "type": "BDH",
            "sec": self.get_argument("sec"),
            "field": self.get_argument("field"),
            "start": self.get_argument("start"),
            "end": self.get_argument("end"),
            "adjusted": self.get_argument("adjusted", True),
            "overrides": self.get_argument("overrides", "null"),
        }

        rt = self.get_argument("RT", False)

        status, resp = yield self.application.process_request(request, rt)

        self.set_status(status)
        self.write(resp)


###############################################################################
class BbgUniqueIdHandler(tornado.web.RequestHandler):
    # -------------------------------------------------------------------------
    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def get(self):
        self.set_header("Content-Type", "application/json")
        self.set_header("Access-Control-Allow-Origin", "*")

        reqtype = self.get_argument("type")
        real_time = self.get_argument("real-time")

        if reqtype == "BDP":
            request = {
                "type": "BDP",
                "sec": self.get_argument("sec"),
                "field": self.get_argument("field"),
                "overrides": self.get_argument("overrides", None),
            }
        elif reqtype == "BDH":
            request = {
                "type": "BDH",
                "sec": self.get_argument("sec"),
                "field": self.get_argument("field"),
                "start": self.get_argument("start"),
                "end": self.get_argument("end"),
                "adjusted": self.get_argument("adjusted", True),
                "overrides": self.get_argument("overrides", None),
            }

        req_uid = request_to_uid(request, real_time)

        self.set_status(200)
        self.write(req_uid)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import GetObj
from ..exceptions import FieldError

__all__ = [
    "get_bbg_ticker",
    "get_bbg_composite",
    "ticker_active",
    "MissingBbgCode",
    "MissingBbgTicker"
]

INACTIVE = {"NONE", "INEG"}


###############################################################################
class MissingBbgCode(Exception):
    pass


###############################################################################
class MissingBbgTicker(Exception):
    pass


# -----------------------------------------------------------------------------
def get_bbg_ticker(sec):
    """
    Description:
        Return the Bloomberg ticker for a given security (if available).
    Inputs:
        sec - the security name
    Returns:
        A string.
    """
    sec = GetObj(sec)
    obj_type = sec.ObjType

    try:
        bbg_ticker = sec.Ticker(platform="Bloomberg")
    except KeyError:
        raise MissingBbgTicker("missing Bloomberg "
                               "ticker for {0!s}".format(sec))

    if obj_type in ("EquityAsset", "DepositaryReceipt"):
        exchange = GetObj(sec.Exchange)

        try:
            bbg_exch_code = exchange.Code(platform="Bloomberg")
        except KeyError:
            raise MissingBbgCode("missing Bloomberg "
                                 "code for {0!s}".format(exchange))

        return "{0:s} {1:s} EQUITY".format(bbg_ticker, bbg_exch_code)

    elif obj_type in ("EquityIndex", "EquityIndexCnt"):
        return "{0:s} INDEX".format(bbg_ticker)

    elif obj_type in ("CommodAsset", "CommodCnt"):
        return "{0:s} COMDTY".format(bbg_ticker)

    elif obj_type in {"BbgSecurity", "CurrencyCross"}:
        return bbg_ticker

    else:
        raise NotImplementedError("function not implemented "
                                  "for security {0:s}".format(sec))


# -----------------------------------------------------------------------------
def get_bbg_composite(bbg_ticker, bbg_clt):
    """
    Description:
        Return the Bloomberg composite ticker for a given ticker, if
        available
    Inputs:
        bbg_ticker - the Bloomberg ticker.
        bbg_clt    - an instance of the bloomberg datafeed client
    Returns:
        A string.
    """
    try:
        ticker = bbg_clt.BDP(bbg_ticker, "EU_COMPOSITE_TICKER")
    except FieldError:
        return bbg_ticker
    else:
        if ticker is None:
            return bbg_ticker
        else:
            return ticker


# -----------------------------------------------------------------------------
def ticker_active(bbg_ticker, bbg_clt):
    """
    Description:
        Check if a Bloomberg ticker is currently active (as opposed to
        inactive, expired, etc).
    Inputs:
        bbg_ticker - the Bloomberg ticker.
        bbg_clt    - an instance of the bloomberg datafeed client
    Returns:
        A tuple (active, status), where:
            active - a bool, True if the ticker is inactive
            status - a string, current market status

    The possible values for stats are:

        ABAL - auction order book balancing
        AUCT - auction call phase
        CAUC - continuous auction pre-call or call phase dependent on market
               makers or specialists
        CLOS - market closed
        HALT - temporary halt from trading
        HLTA - temporary halt and auction call phase
        INEG - ineligible to trade, e.g. pending listing, delisted, private
               company, expired, called, etc.
        MOC  - market-on-close order collection for eventual execution at the
               determined closing price
        MOCB - market-on-close order balancing to try and maximize execution at
               the determined closing price
        NONE - the current status of the security is undefined
        NOTR - not presently available for trading
        OUT  - trading in an out-of-hours trading phase
        POST - post-close session
        PREO - pre-open session
        PRPO - pre-open/post-close session
        NOTR - not presently available for trading
        OUT  - trading in an out-of-hours trading phase
        POST - post-close session
        PREO - pre-open session
        PRPO - pre-open/post-close session
        SUSP - presently suspended from trading
        TALC - trade-at-last or trade-at-close trading phase
        TMCB - actively trading, but with market-on-close order balancing
               operating in parallel
        TMOC - actively trading, but with market-on-close order collection
               operating in parallel
        TRAD - actively trading, e.g. continuous trading, mandatory quotation
               period, etc.
        TRAU - actively trading, with an auction call running in parallel
        UNCR - auction uncrossing phase

    If you are looking to track the status of a security in simplistic terms,
    then you make wish to make use of the following generalizations:

    - A security has an unknown status when it returns a value of NONE.
    - A security is ineligible to trade when it returns a value of INEG.
    - A security is closed for trading when it returns a value of CLOS (if
      QUOTE_STATUS/RT_QUOTE_STATUS is also equal to 'N').
    - A security is available and open for trading (e.g. continuous trading)
      when it returns a value of CAUC, TMCB, TMOC, TRAD or TRAU.
    - A security is available for trade-at-last or trade-at-close order
      executions only when it returns a value of TALC.
    - A security is available for out-of-hours trading when it returns a value
      of OUT.
    - A security is in an auction call when it returns a value of AUCT, HLTA
      or TRAU. If IN_AUCTION/IN_AUCTION_RT returns a value of 'NO' then the
      order executions (the uncrossing) has already taken place at the end of
      that auction call.
    - A security is executing orders from an auction call when it returns a
      value of UNCR (not supported on all markets).
    - A security is offering an opportunity to offset an outstanding imbalance
      from an auction call when it returns a value of ABAL.
    - A security is operating a distinct market-on-close (MOC) order book when
      it returns a value of TMOC or MOC.
    - A security is offering an opportunity to offset an outstanding imbalance
      on a market-on-close.
    - A security is operating a distinct market-on-close (MOC) order book when
      it returns a value of TMOC or MOC.
    - A security is offering an opportunity to offset an outstanding imbalance
      on a market-on-close (MOC) book when it returns a value of MOCB or TMCB.
    - A security is halted or suspended from trading when it returns a value
      of HALT, HLTA or SUSP. If the value is HLTA then the security is
      participating in an auction call to facilitate a resumption of trading.
    - A security is temporarily unavailable for trading - but not halted or
      suspended - when it returns a value of NOTR.
    - A security is in a pre-trading or post-trading phase (auction calls
      excepted), but with no trading via the primary market mechanism (e.g.
      continuous trading), when it returns a value of CLOS (if
      QUOTE_STATUS/RT_QUOTE_STATUS is also equal to 'Y'), POST, PREO or POST.

    Note that there is a distinction between what we understand to be most
    likely temporary halts (HALT and HLTA), and what we understand may be
    longer term suspensions (SUSP). All three values nevertheless represent an
    unscheduled interruption to trading.
    """
    ticker = bbg_ticker.upper()
    status = bbg_clt.BDP(ticker, "SIMP_SEC_STATUS")
    return status not in INACTIVE, status



###############################################################################
#
#   Copyright: (c) 2015-2018 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import Date, Curve, HlocvCurve, GCurve

from ..exceptions import SecurityError, FieldError
from ..exceptions import DatafeedError, DatafeedFatal

import datetime
import blpapi

msg_err_fmt = "({0!s}, {1!s}, {2!s})"
sec_err_fmt = "({0!s}, {1!s}, {2!s})"
fld_err_fmt = "{0!s}, {1!s}, {2!s}, {3!s}"

SECURITY_DATA = blpapi.Name("securityData")
SECURITY = blpapi.Name("security")
CATEGORY = blpapi.Name("category")
SUBCATEGORY = blpapi.Name("subcategory")
MESSAGE = blpapi.Name("message")
RESPONSE_ERROR = blpapi.Name("responseError")
SEC_ERROR = blpapi.Name("securityError")
FIELD_DATA = blpapi.Name("fieldData")
FIELD_EXCEPTIONS = blpapi.Name("fieldExceptions")
FIELD_ID = blpapi.Name("fieldId")
ERROR_INFO = blpapi.Name("errorInfo")

bbg_datatypes = blpapi.datatype.DataType()

VANILLA_TYPES = {
    bbg_datatypes.BOOL,
    bbg_datatypes.CHAR,
    bbg_datatypes.BYTE,
    bbg_datatypes.INT32,
    bbg_datatypes.INT64,
    bbg_datatypes.FLOAT32,
    bbg_datatypes.FLOAT64,
}
bbg_datatypes.VANILLA_TYPES = VANILLA_TYPES


# -----------------------------------------------------------------------------
def get_sequence_value(node):
    """
    Convert an element with DataType Sequence to a list of dictionaries.
    Note this may be a naive implementation as it assumes that bulk data is
    always a table.
    """
    data, cols = [], []
    for i in range(node.numValues()):
        row = node.getValue(i)
        # --- get the ordered cols and assume they are constant
        if i == 0:
            cols = [str(row.getElement(k).name())
                    for k in range(row.numElements())]

        values = []
        for cidx in range(row.numElements()):
            elm = row.getElement(cidx)
            if elm.numValues():
                values.append(get_value(elm))
            else:
                values.append(None)

        data.append(dict(zip(cols, values)))

    return data


# -----------------------------------------------------------------------------
def get_value(ele):
    """
    Convert the specified element as a python value.
    """
    dtype = ele.datatype()
    if dtype in bbg_datatypes.VANILLA_TYPES:
        return ele.getValue()

    elif dtype == bbg_datatypes.STRING:
        return str(ele.getValue())

    elif dtype == bbg_datatypes.DATE:
        v = ele.getValue()
        if v is None:
            return None
        else:
            return Date(year=v.year, month=v.month, day=v.day)

    elif dtype == bbg_datatypes.TIME:
        v = ele.getValue()
        if v is None:
            return None
        else:
            return datetime.time(hour=v.hour, minute=v.minute, second=v.second)

    elif dtype == bbg_datatypes.DATETIME:
        v = ele.getValue()
        if v is None:
            return None
        else:
            return Date(year=v.year, month=v.month, day=v.day,
                        hour=v.hour, minute=v.minute, second=v.second)

    elif dtype == bbg_datatypes.ENUMERATION:
        raise NotImplementedError("ENUMERATION "
                                  "data type not implemented")

    elif dtype == bbg_datatypes.SEQUENCE:
        return get_sequence_value(ele)

    elif dtype == bbg_datatypes.CHOICE:
        raise NotImplementedError("CHOICE data type not implemented")

    else:
        raise NotImplementedError("Unexpected data type {0:s}. "
                                  "Check documentation".format(dtype))


# -----------------------------------------------------------------------------
def check_for_response_errors(resp):
    if resp.hasElement(RESPONSE_ERROR):
        error = resp.getElement(RESPONSE_ERROR)
        cat = error.getElementAsString(CATEGORY)
        subcat = error.getElementAsString(SUBCATEGORY)
        msg = error.getElementAsString(MESSAGE)
        raise DatafeedError(msg_err_fmt.format(cat, subcat, msg))


# -----------------------------------------------------------------------------
def check_for_security_errors(sec_data):
    if sec_data.hasElement(SEC_ERROR):
        error = sec_data.getElement(SEC_ERROR)
        sec = sec_data.getElementAsString(SECURITY)
        cat = error.getElementAsString(CATEGORY)
        msg = error.getElementAsString(MESSAGE)
        raise SecurityError(sec_err_fmt.format(sec, cat, msg))


# -----------------------------------------------------------------------------
def check_for_field_errors(sec_data):
    field_exceptions = list(sec_data.getElement(FIELD_EXCEPTIONS).values())
    assert len(field_exceptions) <= 1
    for field_exception in field_exceptions:
        info = field_exception.getElement(ERROR_INFO)
        field = field_exception.getElementAsString(FIELD_ID)
        sec = sec_data.getElementAsString(SECURITY)
        cat = info.getElementAsString(CATEGORY)
        msg = info.getElementAsString(MESSAGE)
        raise FieldError(fld_err_fmt.format(sec, field, cat, msg))


# -----------------------------------------------------------------------------
def test_bbg_data_service():
    session = blpapi.Session()

    try:
        active = session.start()
    except:
        return False

    try:
        active = session.openService("//blp/refdata")
    except:
        return False
    finally:
        session.stop()

    return active


# -----------------------------------------------------------------------------
def get_session():
    session = blpapi.Session()

    if not session.start():
        raise DatafeedFatal("failed to start session")

    return session


# -----------------------------------------------------------------------------
def get_service(session, service_name):
    if not session.openService(service_name):
        raise DatafeedFatal("failed "
            "to open '{):s}' service".format(service_name))

    return session.getService(service_name)


# -----------------------------------------------------------------------------
def bdp_request(service, securities, fields, overrides=None):
    if overrides is None:
        overrides = {}

    request = service.createRequest("ReferenceDataRequest")

    for sec in securities:
        request.append("securities", sec)
    for field in fields:
        request.append("fields", field)

    for override, value in overrides.items():
        ovd = request.getElement("overrides").appendElement()
        ovd.setElement("fieldId", override)
        ovd.setElement("value", value)

    return request


# -----------------------------------------------------------------------------
def bdp_generic_handler(resp):
    sec_data_array = list(resp.getElement(SECURITY_DATA).values())

    resp = {}
    for sec_data in sec_data_array:
        check_for_security_errors(sec_data)
        check_for_field_errors(sec_data)
        
        sec_name = str(sec_data.getElement("security").getValue())
        resp[sec_name] = {}

        field_data = sec_data.getElement(FIELD_DATA)
        for field in field_data.elements():
            field_name = str(field.name())
            resp[sec_name][field_name] = get_value(field)

    return resp


# -----------------------------------------------------------------------------
def bdp_one_sec_one_field_handler(resp):
    sec_data_array = list(resp.getElement(SECURITY_DATA).values())
    assert len(sec_data_array) == 1

    for sec_data in sec_data_array:
        check_for_security_errors(sec_data)
        check_for_field_errors(sec_data)

        field_data = sec_data.getElement(FIELD_DATA)
        if field_data.numElements():
            return get_value(field_data.getElement(0))
        else:
            return None


# -----------------------------------------------------------------------------
def process_request(session, request, response_handler, timeout=500):
    cid = blpapi.CorrelationId()
    queue = blpapi.EventQueue()
    session.sendRequest(request, correlationId=cid, eventQueue=queue)

    while True:
        ev = queue.nextEvent(timeout)

        # --- request timed out, raise exception
        if ev.eventType() == blpapi.Event.TIMEOUT:
            raise TimeoutError("request timed out "
                               "after {0!s} milliseconds".format(timeout))

        for resp in ev:
            if not cid in resp.correlationIds():
                continue

            check_for_response_errors(resp)
          
            if not resp.hasElement(SECURITY_DATA):
                raise DatafeedError("response with no "
                                    "security data:\n{0!s}".format(resp))

            return response_handler(resp)

        # --- response completly received before finding any valid data or
        #     errors
        if ev.eventType() == blpapi.Event.RESPONSE:
            sec = request.getElement("securities").values().next()
            field = request.getElement("fields").values().next()
            msgfmt = ("query for ({0!s}, {1!s}) "
                          "didn't return a valid response: {2!s}")
            raise DatafeedError(msgfmt.format(field, sec, resp))


###############################################################################
class bdp_client():
    # -------------------------------------------------------------------------
    def __init__(self):
        self.session = get_session()
        self.service = get_service(self.session, "//blp/refdata")

    # -------------------------------------------------------------------------
    def fetch(self, sec, field, overrides=None, timeout=500):
        request = bdp_request(self.service, [sec], [field], overrides)
        resp = process_request(self.session, request,
                               bdp_one_sec_one_field_handler, timeout)
        return resp

    # -------------------------------------------------------------------------
    def fetch_many(self, securities, fields, overrides=None, timeout=500):
        request = bdp_request(self.service, securities, fields, overrides)
        resp = process_request(self.session, request,
                               bdp_generic_handler, timeout)
        return resp

    # -------------------------------------------------------------------------
    def __del__(self):
        self.session.stop()


# -----------------------------------------------------------------------------
def BbgBDP(sec, field, overrides=None, timeout=500):
    """
    Description:
        Executes a blocking BDP request.
    Inputs:
        sec       - a valid bloomberg identifier
        field     - a valid bloomberg field for the above identifier
        overrides - an optional dictionary of overrides for the above
                    identifier
        timeout   - timeout interval in milliseconds. A value of zero means
                    wait forever.
    Returns:
        A dictionary or raises TimeoutError if the request times out.
    """
    clt = bdp_client()
    try:
        return clt.fetch(sec, field, overrides, timeout)
    finally:
        del clt


# -----------------------------------------------------------------------------
def BbgBDH(sec, field, start, end,
           adjusted=True, overrides=None, timeout=500, **kwds):
    """
    Description:
        Executes a blocking BDH request.
    Inputs:
        sec       - a valid bloomberg identifier
        field     - a valid bloomberg field for the above identifier. "HLOCV"
                    is a special acceptable field type.
        start     - start date, in yyyymmdd format
        end       - end date, in yyyymmdd format
        adjusted  - True for dividend/split adjusted data
        overrides - an optional dictionary of overrides for the above
                    identifier
        timeout   - timeout interval in milliseconds. A value of zero means
                    wait forever.
    Returns:
        Depending on fields, a Curve or HlocvCurve or GCurve or raises
        TimeoutError if the request times out.
    """
    if field == "HLOCV":
        fields = ["PX_HIGH", "PX_LOW", "PX_OPEN", "PX_LAST", "VOLUME"]
        is_hlocv = True
    else:
        is_hlocv = False
        fields = [field]

    # --- fildnames is used to iterate over fieldData element
    field_names = ["date"] + fields

    if overrides is None:
        overrides = {}

    session = blpapi.Session()

    if not session.start():
        raise DatafeedFatal("failed to start session")

    try:
        if not session.openService("//blp/refdata"):
            raise DatafeedFatal("failed to open '//blp/refdata' service")

        service = session.getService("//blp/refdata")
        request = service.createRequest("HistoricalDataRequest")

        request.append("securities", sec)
        for fld in fields:
            request.append("fields", fld)

        request.set("periodicitySelection", "DAILY")
        request.set("startDate", start)
        request.set("endDate", end)
        request.set("adjustmentNormal", adjusted)
        request.set("adjustmentAbnormal", adjusted)
        request.set("adjustmentSplit", adjusted)

        for override, value in overrides.items():
            ovd = request.getElement("overrides").appendElement()
            ovd.setElement("fieldId", override)
            ovd.setElement("value", value)

        cid = blpapi.CorrelationId(1)
        queue = blpapi.EventQueue()
        session.sendRequest(request, correlationId=cid, eventQueue=queue)

        data = []
        while True:
            ev = queue.nextEvent(timeout)

            # --- request timed out, raise exception
            if ev.eventType() == blpapi.Event.TIMEOUT:
                raise TimeoutError("request timed out "
                                   "after {0!s} milliseconds".format(timeout))

            for resp in ev:
                if not cid in resp.correlationIds():
                    continue

                check_for_response_errors(resp)
              
                if not resp.hasElement(SECURITY_DATA):
                    raise DatafeedError("response with no "
                                        "security data:\n{0!s}".format(resp))

                sec_data = resp.getElement(SECURITY_DATA)

                check_for_security_errors(sec_data)
                check_for_field_errors(sec_data)

                field_data_array = sec_data.getElement(FIELD_DATA)
                for fieldData in field_data_array.values():
                    knot = [None, {}]
                    for field_name in field_names:
                        if fieldData.hasElement(field_name):
                            value = get_value(fieldData.getElement(field_name))
                        else:
                            value = None

                        if field_name == "date":
                            knot[0] = value
                        else:
                            knot[1][field_name] = value

                    data.append(knot)

            # --- response completly received, so we could exit
            if ev.eventType() == blpapi.Event.RESPONSE:
                break

    finally:
        session.stop()

    if is_hlocv:
        if len(data):
            # --- remove knots without a close value (should be extremely rare)
            knots = [(d, [v[fld] for fld in fields])
                     for d, v in data if v["PX_LAST"] is not None]
            return HlocvCurve([d for d, v in knots], [v for d, v in knots])
        else:
            return HlocvCurve()

    elif len(fields) == 1:
        if len(data):
            fld = fields[0]
            try:
                float(data[0][1][fld])
            except ValueError:
                return data
            else:
                knots = [(d, v[fld]) for d, v in data]
                return Curve([d for d, v in knots], [v for d, v in knots])
        else:
            return Curve()
    else:
        return GCurve([d for d, v in data], [v for d, v in data])



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from ..clients import DataClient
from ..exceptions import NoDataserverAvailable

from onyx.core.utils.logging import setup_logging_from_json_config

import logging
import time
import argh


# -----------------------------------------------------------------------------
def run(ticker="USDEUR CURNCY",
        keepalive_interval=120, logging_config_file=None):
    setup_logging_from_json_config(logging_config_file)
    logger = logging.getLogger(__name__)
    clt = DataClient()
    while True:
        try:
            clt.BDP(ticker, "PX_LAST")
        except NoDataserverAvailable as err:
            logger.info(err)
        except:
            logger.error("connection to datafeed "
                         "server is not working", exc_info=True)
        else:
            logger.info("conection to datafeed server is alive")

        time.sleep(keepalive_interval)


# -----------------------------------------------------------------------------
def main():
    argh.dispatch_command(run)


# -----------------------------------------------------------------------------
#  for interactive use
if __name__ == "__main__":
    run()



###############################################################################
#
#   Copyright: (c) 2017 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core.database.ufo_base import custom_encoder, custom_decoder

import json
import hashlib

REQ_FIELDS = ("sec", "field", "start", "end",
              "adjusted", "overrides", "type", "real-time")


# -----------------------------------------------------------------------------
def encode_message(msg):
    encoded_msg = json.dumps(msg, separators=(",", ":"), cls=custom_encoder)
    return bytes(encoded_msg + "\n", "utf-8")


# -----------------------------------------------------------------------------
def decode_message(msg):
    return json.loads(msg.decode("utf-8"), cls=custom_decoder)


# -----------------------------------------------------------------------------
def request_to_uid(request, real_time):
    # --- serialize request making sure we always get the same key for the
    #     same request (NB: the request is a dictionary which means
    #     ordering is not guaranteed).
    req_vls = ["({0!s}".format(request.get(k, None)) for k in REQ_FIELDS]
    req_vls.append("{0!s}".format(real_time))
    return hashlib.sha1("".join(req_vls).encode("utf-8")).hexdigest()



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import Date
from ..exceptions import DatafeedError

import win32com.client
import collections
import datetime

__all__ = []

security_error_attrs = ["security", "source", "code",
                        "category", "message", "subcategory"]
security_error = collections.namedtuple("SecurityError", security_error_attrs)

field_error_attrs = ["security", "field", "source",
                     "code", "category", "message", "subcategory"]
field_error = collections.namedtuple("FieldError", field_error_attrs)


# -----------------------------------------------------------------------------
def security_iter(nodearr):
    """
    Provide a security data iterator by returning a tuple of
    (Element, SecurityError) which are mutually exclusive
    """
    if nodearr.Name != "securityData":
        raise RuntimeError("nodearr.Name != 'securityData'")
    if not nodearr.IsArray:
        raise RuntimeError("nodearr.IsArray is False")

    for i in range(nodearr.NumValues):
        node = nodearr.GetValue(i)
        err = get_security_error(node)
        yield (None, err) if err else (node, None)


# -----------------------------------------------------------------------------
def message_iter(evt):
    """
    Provide a message iterator which checks for a response error prior
    to returning
    """
    msg_iter = evt.CreateMessageIterator()
    while msg_iter.Next():
        msg = msg_iter.Message
        if msg.AsElement.HasElement("responseError"):
            raise DatafeedError(msg.AsElement.GetValue("message"))
        yield msg


# -----------------------------------------------------------------------------
def get_sequence_value(node):
    """
    Convert an element with DataType Sequence to a list of dictionaries.
    Note this may be a naive implementation as it assumes that bulk data is
    always a table.
    """
    data, cols = [], []
    for i in range(node.NumValues):
        row = node.GetValue(i)
        # --- get the ordered cols and assume they are constant
        if i == 0:
            cols = [str(row.GetElement(k).Name)
                    for k in range(row.NumElements)]

        values = []
        for cidx in range(row.NumElements):
            elm = row.GetElement(cidx)
            values.append(as_value(elm))

        data.append(dict(zip(cols, values)))

    return data


# -----------------------------------------------------------------------------
def as_value(ele):
    """
    Convert the specified element as a python value.
    """
    dtype = ele.Datatype
    if dtype in (1, 2, 3, 4, 5, 6, 7, 9, 12):
        # --- BOOL, CHAR, BYTE, INT32, INT64, FLOAT32, FLOAT64,
        #     BYTEARRAY, DECIMAL
        return ele.Value
    elif dtype == 8:  # string
        return str(ele.Value)
    elif dtype == 10:  # date
        v = ele.Value
        if v is None:
            return None
        else:
            return Date(year=v.year, month=v.month, day=v.day)
    elif dtype == 11:  # time
        v = ele.Value
        if v is None:
            return None
        else:
            return datetime.time(hour=v.hour, minute=v.minute, second=v.second)
    elif dtype == 13:  # datetime
        v = ele.Value
        if v is None:
            return None
        else:
            return Date(year=v.year, month=v.month, day=v.day,
                        hour=v.hour, minute=v.minute, second=v.second)
    elif dtype == 14:  # Enumeration
        raise NotImplementedError("ENUMERATION "
                                  "data type needs implemented")
    elif dtype == 15:  # SEQUENCE
        return get_sequence_value(ele)
    elif dtype == 16:  # Choice
        raise NotImplementedError("CHOICE data type needs implemented")
    else:
        raise NotImplementedError("Unexpected data type {0:s}. "
                                  "Check documentation".format(dtype))


# -----------------------------------------------------------------------------
def get_child_value(parent, name):
    """
    Return the value of the child element with name in the parent Element
    """
    if parent.HasElement(name):
        return as_value(parent.GetElement(name))
    else:
        return None


# -----------------------------------------------------------------------------
def get_child_values(parent, names):
    """
    Return a list of values for the specified child fields. If field not in
    Element then replace with nan.
    """
    vals = []
    for name in names:
        if parent.HasElement(name):
            vals.append(as_value(parent.GetElement(name)))
        else:
            vals.append(None)
    return vals


# -----------------------------------------------------------------------------
def as_security_error(node, secid):
    """
    Convert the securityError element to a SecurityError.
    """
    if node.Name != "securityError":
        raise RuntimeError("node.Name != 'securityError'")

    src = get_child_value(node, "source")
    code = get_child_value(node, "code")
    cat = get_child_value(node, "category")
    msg = get_child_value(node, "message")
    subcat = get_child_value(node, "subcategory")
    return security_error(security=secid, source=src, code=code,
                          category=cat, message=msg, subcategory=subcat)


# -----------------------------------------------------------------------------
def as_field_error(node, secid):
    """
    Convert a fieldExceptions element to a FieldError or FieldError array.
    """
    if node.Name != "fieldExceptions":
        raise RuntimeError("node.Name != 'fieldExceptions'")

    if node.IsArray:
        return [as_field_error(node.GetValue(k),
                               secid) for k in range(node.NumValues)]
    else:
        fld = get_child_value(node, "fieldId")
        info = node.GetElement("errorInfo")
        src = get_child_value(info, "source")
        code = get_child_value(info, "code")
        cat = get_child_value(info, "category")
        msg = get_child_value(info, "message")
        subcat = get_child_value(info, "subcategory")
        return field_error(security=secid, field=fld, source=src,
                           code=code, category=cat, message=msg,
                           subcategory=subcat)


# -----------------------------------------------------------------------------
def get_security_error(node):
    """
    Return a SecurityError if the specified securityData element has one,
    else return None
    """
    if node.Name != "securityData":
        raise RuntimeError("node.Name != 'securityData'")
    if node.IsArray:
        raise RuntimeError("node.IsArray is True")

    if node.HasElement("securityError"):
        secid = get_child_value(node, "security")
        element = node.GetElement("securityError")
        return as_security_error(element, secid)
    else:
        return None


# -----------------------------------------------------------------------------
def get_field_errors(node):
    """
    Return a list of FieldErrors if the specified securityData element has
    field errors
    """
    if node.Name != "securityData":
        raise RuntimeError("node.Name != 'securityData'")
    if node.IsArray:
        raise RuntimeError("node.IsArray is True")

    nodearr = node.GetElement("fieldExceptions")
    if nodearr.NumValues > 0:
        secid = get_child_value(node, "security")
        return as_field_error(nodearr, secid)
    else:
        return None


# -----------------------------------------------------------------------------
def debug_event(evt):
    print("unhandled event: {0:s}".format(evt.EventType))
    if evt.EventType in [win32com.client.constants.RESPONSE,
                         win32com.client.constants.PARTIAL_RESPONSE]:
        print("messages:")
        for msg in message_iter(evt):
            print(msg.Print)



###############################################################################
#
#   Copyright: (c) 2017 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import load_system_configuration
from onyx.core import Date
from onyx.core import CurveField, HlocvCurveField, GCurveField

from .utils import encode_message, decode_message

from .exceptions import DatafeedError, DatafeedFatal, SecurityError, FieldError
from .bloomberg import blp_api as bbg_api

from concurrent import futures

import tornado.ioloop
import tornado.tcpserver
import tornado.httpclient
import tornado.gen
import tornado.netutil

import time
import json
import datetime
import socket
import urllib
import logging

__all__ = ["DataServer"]

API_ERRORS = DatafeedError, SecurityError, FieldError, NotImplementedError
REREGISTER = 60000*2  # check registration every 2 minutes


# -----------------------------------------------------------------------------
def cycle(elements):
    while True:
        for element in elements:
            yield element


# -----------------------------------------------------------------------------
def bdp_request(bdp_clt, timeout, request):
    try:
        data = bdp_clt.fetch(timeout=timeout, **request)
        return {
            "type": type(data).__name__,
            "payload": data,
        }
    except API_ERRORS as err:
        return {
            "type": type(err).__name__,
            "payload": str(err),
        }


# -----------------------------------------------------------------------------
def bdh_request(timeout, request):
    try:
        data = bbg_api.BbgBDH(timeout=timeout, **request)
        return {
            "type": type(data).__name__,
            "payload": data,
        }
    except API_ERRORS as err:
        return {
            "type": type(err).__name__,
            "payload": str(err),
        }


# -----------------------------------------------------------------------------
def process_response(resp):
    dtype = resp["type"]
    if dtype == "Curve":
        resp["payload"] = CurveField.to_json(None, resp["payload"])
    elif dtype == "HlocvCurve":
        resp["payload"] = HlocvCurveField.to_json(None, resp["payload"])
    elif dtype == "GCurve":
        resp["payload"] = GCurveField.to_json(None, resp["payload"])
    return resp


###############################################################################
class DataServer(tornado.tcpserver.TCPServer):
    # -------------------------------------------------------------------------
    def __init__(self, logger=None, nthreads=10, blp_timeout=1, tcp_timeout=5):
        super().__init__()
        self.logger = logger or logging.getLogger(__name__)
        self.tpool = futures.ThreadPoolExecutor(max_workers=nthreads)
        self.blp_timeout = blp_timeout * 1000  # seconds to milliseconds
        self.tcp_timeout = datetime.timedelta(seconds=tcp_timeout)
        self.bdp_clients = cycle([bbg_api.bdp_client()
                                  for thread in range(nthreads)])

    # -------------------------------------------------------------------------
    def start(self, router_addr=None, router_port=None, stop_at=None):
        config = load_system_configuration()
        router_addr = router_addr or config.get("datafeed", "router_address")
        router_port = router_port or config.getint("datafeed", "router_port")

        # --- connect to a random port from the available pool
        sockets = tornado.netutil.bind_sockets(0, address="")
        self.add_sockets(sockets)

        # --- get the port and add a callback to notify periodically that
        #     this server is available
        port = sockets[0].getsockname()[1]

        # --- create the request object used to notify the router that this
        #     server is available
        self.subreq = tornado.httpclient.HTTPRequest(
            "http://{0!s}:{1!s}/register/".format(router_addr, router_port),
            method="PUT",
            headers={"Content-Type": "application/x-www-form-urlencoded"},
            body=urllib.parse.urlencode({
                "address": socket.gethostbyname(socket.gethostname()),
                "port": port}),
            request_timeout=self.tcp_timeout.seconds
        )

        ioloop = tornado.ioloop.IOLoop.current()
        ioloop.add_callback(self.subscribe_to_router)
        tornado.ioloop.PeriodicCallback(
            self.subscribe_to_router, REREGISTER, ioloop).start()

        if stop_at is not None:
            seconds = (stop_at - Date.now()).seconds
            ioloop.call_later(seconds, lambda: self.stop_server())

        try:
            ioloop.start()
        except KeyboardInterrupt:
            self.stop_server()

    # -------------------------------------------------------------------------
    def stop_server(self):
        # --- first stop the threadpool
        self.tpool.shutdown(wait=True)
        # --- then stop the tornado event loop
        ioloop = tornado.ioloop.IOLoop.current()
        ioloop.stop()
        self.logger.info("shutting down server")

    # -------------------------------------------------------------------------
    @tornado.gen.coroutine
    def with_timeout(self, future):
        generator = yield tornado.gen.with_timeout(self.tcp_timeout, future)
        return generator

    # -------------------------------------------------------------------------
    @tornado.gen.coroutine
    def handle_stream(self, stream, address):
        # --- use time.time to monitor time needed to process request
        t_start = time.time()

        self.logger.debug("connection received from {0!s}".format(address))

        try:
            request = yield self.with_timeout(stream.read_until(b"\n"))
        except (tornado.iostream.StreamClosedError, tornado.gen.TimeoutError):
            self.logger.error("couldn't read "
                              "request from {0!s}".format(address))
            return

        request = decode_message(request)
        self.logger.debug("processing request {0!s}".format(request))

        request_type = request.pop("type")

        # --- overrides, which are sent by clients as json strings, are decoded
        #     here 'in-place'
        request["overrides"] = json.loads(request["overrides"])

        # --- try processing the request for a maximum of 5 times using
        #     geometrically increasing timeouts
        timeout = self.blp_timeout
        for k in range(5):
            self.logger.debug("timeout set to {0:d}".format(timeout))
            try:
                if request_type == "BDP":
                    # --- pull a bbg-bdp client from the clients pool
                    clt = next(self.bdp_clients)
                    resp = yield self.tpool.submit(bdp_request,
                                                   clt, timeout, request)
                else:
                    resp = yield self.tpool.submit(bdh_request,
                                                   timeout, request)
                break
            except TimeoutError:
                timeout *= 2
                self.logger.info("request {0!s} "
                    "timeout increased to {1:f}ms".format(request, timeout))
            except DatafeedFatal as err:
                # --- unrecoverable bloomberg error: don't send a reply so that
                #     the datafeed router knows this server is currently unable
                #     to respond to requests.
                self.logger.error(err, exc_info=True)
                return

        else:
            timeout /= 2
            resp = {
                "type": "TimeoutError",
                "payload": ("request timed out after 5 attempts with a final "
                            "timeout of {0!s} milliseconds").format(timeout),
            }

        resp = process_response(resp)
        resp = encode_message(resp)

        try:
            yield self.with_timeout(stream.write(resp))
        except (tornado.gen.TimeoutError, tornado.iostream.StreamClosedError):
            self.logger.error("couldn't "
                "send resp to {0!s}".format(address))
            return

        time_total = (time.time() - t_start)*1000.0
        self.logger.info("request {0!s} "
            "processed in {1:f}ms".format(request, time_total))

    # -------------------------------------------------------------------------
    @tornado.gen.coroutine
    def subscribe_to_router(self):
        # --- register only if bloomberg service is available
        active = bbg_api.test_bbg_data_service()
        if not active:
            self.logger.info("couldn't register server "
                "with router: bloomberg service currently unavailable")
            return

        self.logger.debug("trying to register "
            "with datafeed router on {0!s}".format(self.subreq.url))

        client = tornado.httpclient.AsyncHTTPClient()
        try:
            res = yield client.fetch(self.subreq)
        except ConnectionRefusedError:
            self.logger.info("couldn't register with datafeed router "
                "on {0!s}: connection refused".format(self.subreq.url))
            return
        except tornado.httpclient.HTTPError as err:
            self.logger.info("couldn't register with "
                "datafeed router on {0!s}: {1!s}".format(self.subreq.url, err))
            return

        if res.code == 201:
            self.logger.info("router accepted subscription")
        elif res.code == 205:
            # --- server already registered with router
            pass
        else:
            raise res.error



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import Date, DateOffset
from ..bloomberg.com_api import IntrdayBarRequest
from ..bloomberg.com_api import ReferenceDataRequest, HistoricalDataRequest

d = DateOffset(Date.today(), "-4b")
m = DateOffset(Date.today(), "-1m+0J+0b")


def banner(msg):
    print('*' * 25)
    print(msg)
    print('*' * 25)

banner('ReferenceDataRequest: single security, single field')
req = ReferenceDataRequest('msft us equity', 'px_last')
print(req.execute().response)

banner('ReferenceDataRequest: multi-security, multi-field')
req = ReferenceDataRequest(['eurusd curncy', 'msft us equity'],
                           ['px_open', 'px_last'])
print(req.execute().response)

banner('ReferenceDataRequest: single security, '
       'multi-field (with bulk), frame response')
req = ReferenceDataRequest('eurusd curncy', ['px_last', 'fwd_curve'])
req.execute()
print(req.response)

banner('ReferenceDataRequest: multi security, multi-field, bad field')
req = ReferenceDataRequest(['eurusd curncy', 'msft us equity'],
                           ['px_last', 'fwd_curve'], ignore_fld_err=True)
req.execute()
print(req.response)

banner('HistoricalDataRequest: multi security, multi-field, daily data')
req = HistoricalDataRequest(['eurusd curncy', 'msft us equity'],
                            ['px_last', 'px_open'], start=d)
req.execute()
print(req.response)

banner('HistoricalDataRequest: multi security, multi-field, weekly data')
req = HistoricalDataRequest(['eurusd curncy', 'msft us equity'],
                            ['px_last', 'px_open'], start=m, period='WEEKLY')
req.execute()
print(req.response)

banner('IntrdayBarRequest: every hour')
req = IntrdayBarRequest('eurusd curncy', 60, start=d)
req.execute()
print(req.response)

#
# HOW TO
#
# - Retrieve an fx vol surface:
#       BbgReferenceDataRequest('eurusd curncy', 'DFLT_VOL_SURF_MID')
# - Retrieve a fx forward curve:
#       BbgReferenceDataRequest('eurusd curncy', 'FWD_CURVE')
# - Retrieve dividends:
#       BbgReferenceDataRequest('csco us equity',
#                               'BDVD_PR_EX_DTS_DVD_AMTS_W_ANN')
